#!/bin/sh
#rpxqimage

#$1 - file; $2 - size; $3 - offset
read_value() {
    v="$(hexdump -v -n $2 -s $3 -e '1/1 "%02x"' $1)"
    printf $((0x$(eval printf $(printf '%%s%.0s' $(seq 1 $2)) $(printf '"${v:%d:2}" ' $(seq $(($2 * 2 - 2)) -2 0)))))
}

#$1 - file; $2 - option
read_xq_option() {
	if [ "`hexdump -n 4 -e '"%_p"' "$1"`" = "HDR1" ]
	then
		xqv_tmp="$(printf $(mktemp -u | sed -r 's/\.+/_/g'))"
		mkxqimage.elf -V "$1" > "$xqv_tmp"
		uci -q get $xqv_tmp.version.$2
		rm -f "$xqv_tmp" 2>/dev/null
	else
		uci -q get $xqv_tmp.version.$2
	fi
}

#$1 - file; $2 - option; $3 value
write_xq_option() {
	if [ -z "`cat $1 | grep "option $2"`" ]
	then
		echo "   option $2 '$3'" >> "$1"
	elif [ "$(uci -q get $1.version.$2)" != "$3" ]
	then
		sed -i "s/option $2.*/option $2 '$3'/g" "$1"
	fi
}

state="opt"
image=
image_seq=
rpxqimage_ext=
skip_seg=0
skip_patch=0
force=0

for arg in "$@"
do
	case "$state" in
	"opt")
		case "$arg" in
		"-f")
			force=1
			;;
		"-p")
			skip_patch=1
			;;
		"-s")
			skip_seg=1
			;;
		*)
			state="$arg"
			;;
		esac
		continue
		;;
	"-i")
		image="$arg"
		[ "$(dirname $image)" = "." ] && image="$(printf $(pwd))/$(basename $image)"
		;;
	"-e")
		rpxqimage_ext="$arg"
		[ "$(dirname $rpxqimage_ext)" = "." ] && rpxqimage_ext="$(printf $(pwd))/$(basename $rpxqimage_ext)"
		;;
	"-Q")
		image_seq=$(($arg))
		;;
	*)
		echo "Invalid option: $state $arg"
		exit 1
		;;
	esac
	state="opt"
done

[ -f "$image" ] || {
	echo 'Image not found.'
	echo 'Usage: rpxqimage.sh -i IMAGE_PATH [-e SCRIPT_PATH] [-Q NUMBER] [-s] [-f] [-p]'
	echo '-i - path to image file.'
	echo '-s - do not set image as patched (skip "RPXQIMAGE=1" option in xiaoqiang_version).'
	echo '-f - force to repack images with "rpxqimage" segment.'
	echo '-p - do not integrate "rpxqimage" patch.'
	echo '-Q - image sequence number. If not set tries to find sequence number in the image file.'
	echo '-e - path to optional script file.'
	echo '     Takes 2 arguments:'
	echo '        $1 - path to rw upper layer'
	echo '        $2 - path to ro lower layer'
	echo 'Set "xiaoqiang.common.rpxqimage" variable to 0 to disable "rpxqimage.sh" on fw update:'
	echo '     uci set xiaoqiang.common.rpxqimage=0'
	exit 1
}

[ "$(read_xq_option $image 'RPXQIMAGE')" = "1" ] && {
	[ $force -eq 0 ] && exit 0
}

work_dir="$(printf $(printf $(dirname $image)/rpxqimage_workdir | sed -r 's/\.+/_/g'))"
rootfs_rw_dir="$work_dir/merged"
rootfs_ro_dir="$work_dir/lower"
segments="$work_dir/segments"

[ -d "$rootfs_rw_dir" ] && umount -f "$rootfs_rw_dir" 2>/dev/null
[ -d "$rootfs_ro_dir" ] && umount -f "$rootfs_ro_dir" 2>/dev/null
[ -d "$work_dir" ] && { umount -f "$work_dir" 2>/dev/null; rm -rf "$work_dir" 2>/dev/null; }

set -e

[ -z $image_seq ] && {
	image_seq=$((0x$(mkxqimage.elf -rnx "$image" -f root.ubi | head -c28 | tail -c4 | hexdump -n4 -e '1/1 "%02x"')))
}

mkdir -p "$work_dir"
mount -t tmpfs tmpfs "$work_dir"
mkdir -p "$rootfs_ro_dir" "$work_dir/upper" "$rootfs_rw_dir" "$work_dir/work" "$segments"

mkxqimage_cmd="mkxqimage.mkimg.sh -o $image"
mkxqimage_cmd_bkp="mkxqimage.mkimg.sh -o $image"

for word in $(echo $(mkxqimage.elf -c "$image"))
do
	mkxqimage.elf -c "$image" -f "$word" >/dev/null && {
		[ "$word" != "root.ubi" ] && mkxqimage.elf -rnx "$image" -f "$word" > "$segments/$word"
		mkxqimage_cmd="$mkxqimage_cmd -f $segments/$word"
	}
done

dd if="$image" of="$segments/$(basename $image).sig" bs=1 skip=$(read_value "$image" 4 4) 2>/dev/null
mkxqimage_cmd="$mkxqimage_cmd -s $segments/$(basename $image).sig"
mkxqimage_cmd="$mkxqimage_cmd -M $(read_value $image 2 14) -H $(read_value $image 4 16)"
mkxqimage_cmd_bkp="$mkxqimage_cmd"

[ $skip_seg -eq 0 ] && write_xq_option "$segments/xiaoqiang_version" "RPXQIMAGE" 1

ubireader_extract_images.lua -w -o "$work_dir/volumes" "$image"
kernel="$work_dir/volumes/$(ls $work_dir/volumes | grep kernel)"
rootfs="$work_dir/volumes/$(ls $work_dir/volumes | grep rootfs)"

restore_image() {
	set +e
	mkxqimage.elf -rnx "$image" >/dev/null 2>/dev/null || {
		ubinize_cmd='rpxqimage_ubinize.sh -k "$kernel" -r "$rootfs" -o "$segments/root.ubi" -Q "$image_seq"'
		[ -f "$(ls $work_dir/volumes | grep rootfs_data)" ] && ubinize_cmd="$ubinize_cmd -d"
		eval "$ubinize_cmd"
		eval "$mkxqimage_cmd_bkp"
	}
	umount -f "$rootfs_rw_dir"
	umount -f "$rootfs_ro_dir"
	umount -f "$work_dir"
	rm -rf "$work_dir"
}
trap restore_image EXIT

rm -f "$image"

squashfuse "$rootfs" "$rootfs_ro_dir"
mount -t overlay overlay -o lowerdir="$rootfs_ro_dir",upperdir="$work_dir/upper",workdir="$work_dir/work" "$rootfs_rw_dir"

if [ $skip_patch -eq 0 ]
then

[ -z "$(cat $rootfs_rw_dir/bin/flash.sh | grep rpxqimage)" ] && cp -fp "$rootfs_rw_dir/bin/flash.sh" "$rootfs_rw_dir/bin/flash_do_upgrade.sh"
[ -z "$(cat $rootfs_rw_dir/bin/mkxqimage | grep rpxqimage)" ] && cp -fp "$rootfs_rw_dir/bin/mkxqimage" "$rootfs_rw_dir/bin/mkxqimage.elf"

while IFS= read -r file_path
do
	cp -fpP "$file_path" "$rootfs_rw_dir/$file_path"
done << END
/bin/flash.sh
/bin/mkxqimage
/bin/mkxqimage.ubi.sh
/bin/mkxqimage.mkimg.sh
/bin/rpxqimage.sh
/bin/rpxqimage_ubinize.sh
/bin/tar
/usr/sbin/ubireader_extract_images.lua
/usr/sbin/mksquashfs
/usr/sbin/unsquashfs
/usr/bin/squashfuse
/usr/bin/xz
/usr/bin/xzcat
/usr/bin/lzma
/usr/bin/unlzma
/usr/bin/lzcat
/usr/bin/unxz
/usr/lib/libzstd.so.1.4.5
/usr/lib/libzstd.so.1
/usr/lib/libzstd.so
/usr/lib/libulockmgr.so.1.0.1
/usr/lib/libulockmgr.so.1
/usr/lib/libsquashfuse_ll.so.0.0.0
/usr/lib/libsquashfuse_ll.so.0
/usr/lib/libsquashfuse.so.0.0.0
/usr/lib/libsquashfuse.so.0
/usr/lib/liblz4.so.1.9.2
/usr/lib/liblz4.so.1
/usr/lib/libfuseprivate.so.0.0.0
/usr/lib/libfuseprivate.so.0
/usr/lib/libfuse.so.2.9.7
/usr/lib/libfuse.so.2
END

[ $skip_seg -eq 0 ] && write_xq_option "$segments/xiaoqiang_version" "RPXQIMAGE" 1

fi

[ -f "$rpxqimage_ext" ] && { chmod 755 $rpxqimage_ext; $rpxqimage_ext "$rootfs_rw_dir" "$rootfs_ro_dir"; }

mksquashfs "$rootfs_rw_dir" "$rootfs.new" -comp xz -b 256K -no-xattrs -quiet

ubinize_cmd='rpxqimage_ubinize.sh -k "$kernel" -r "$rootfs.new" -o "$segments/root.ubi" -Q "$image_seq"'
[ -f "$(ls $work_dir/volumes | grep rootfs_data)" ] && ubinize_cmd="$ubinize_cmd -d"
eval "$ubinize_cmd"
eval "$mkxqimage_cmd"
